''**********************************
''* VGA_HiRes Text Driver   5/2016 *
''* Author: Werner L. Schneider    *
''* 256 Chars, 8x16 Pixel,         *
''**********************************
''
'' Based on VGA High-Res Text Driver v1.0 from Chip Gracey    Copyright (c) 2006 Parallax, Inc
''
''
CON

' 640 x 480 @ 69Hz settings: 80 x 30 characters

    hp = 640      'horizontal pixels
    vp = 480      'vertical pixels
    hf = 24       'horizontal front porch pixels
    hs = 40       'horizontal sync pixels
    hb = 128      'horizontal back porch pixels
    vf = 9        'vertical front porch lines
    vs = 3        'vertical sync lines
    vb = 28       'vertical back porch lines
    hn = 1        'horizontal normal sync state (0|1)
    vn = 1        'vertical normal sync state (0|1)
    pr = 30       'pixel rate in MHz at 80MHz system clock (5MHz granularity)

' columns and rows

    cols = hp / 8
    rows = vp / 16     


VAR long cog[2]


PUB start(BasePin, ScreenPtr, CursorPtr, FontPtr, Color) : okay | i, j

'' Start VGA driver - starts two COGs
'' returns false if two COGs not available
''
''      BasePin = VGA starting pin (0, 8, 16, 24, etc.)
''
''      ScreenPtr = Pointer to 80x30 words containing Latin-1 codes and colors for
''              each of the 80x30 screen characters. The lower byte of the word
''              contains the Latin-1 code to display. The upper byte contains
''              the foreground colour in bits 11..8 and the background colour in
''              bits 15..12.
''
''              screen word example: %00011111_01000001 = "A", white on blue
''
''      CursorPtr = Pointer to 3 bytes which control the cursor:
''
''              bytes 0,1,2: X, Y, and MODE of cursor 
''
''              X and Y are in terms of screen characters
''              (left-to-right, top-to-bottom)
''
''              MODE uses three bottom bits:
''
''                      %x00 = cursor off
''                      %x01 = cursor on
''                      %x10 = cursor on, blink slow
''                      %x11 = cursor on, blink fast
''                      %0xx = cursor is solid block
''                      %1xx = cursor is underscore
''
''              cursor example: 127, 63, %010 = blinking block in lower-right
''

    ' if driver is already running, stop it
    stop

    ' implant pin settings
    reg_vcfg := $200000FF + (BasePin & %111000) << 6
    i := $FF << (BasePin & %011000)
    j := BasePin & %100000 == 0
    reg_dira := i & j
    reg_dirb := i & !j

    ' implant CNT value to sync COGs to
    sync_cnt := cnt + $10000

    ' implant pointers
    longmove(@screen_base, @ScreenPtr, 3)

    ' implant unique settings and launch first COG
    vf_lines.byte := vf
    vb_lines.byte := vb
    font_part := 1
    cog[1] := cognew(@d0, Color) + 1

    ' allow time for first COG to launch
    waitcnt($2000 + cnt)

    ' differentiate settings and launch second COG
    vf_lines.byte := vf+4
    vb_lines.byte := vb-4
    font_part := 0
    cog[0] := cognew(@d0, Color) + 1

    ' if both COGs launched, return true
    if cog[0] and cog[1]
        return true
    ' else, stop any launched COG and return false
    stop


PUB stop | i

'' Stop VGA driver - frees two COGs

    repeat i from 0 to 1
        if cog[i]
            cogstop(cog[i]~ - 1)


CON

    #1, scanbuff[80], colorbuff[80], scancode[2*80-1+3], maincode   'enumerate COG RAM usage

    main_size = $1F0 - maincode                             'size of main program

    hv_inactive = (hn << 1 + vn) * $0101                    'H,V inactive states


DAT

'*****************************************************
'* Assembly language VGA high-resolution text driver *
'*****************************************************

' This program runs concurrently in two different COGs.
'
' Each COG's program has different values implanted for front-porch lines and
' back-porch lines which surround the vertical sync pulse lines. This allows
' timed interleaving of their active display signals during the visible portion
' of the field scan. Also, they are differentiated so that one COG displays
' even four-line groups while the other COG displays odd four-line groups.
'
' These COGs are launched in the PUB 'start' and are programmed to synchronize
' their PLL-driven video circuits so that they can alternately prepare sets of
' four scan lines and then display them. The COG-to-COG switchover is seemless
' due to two things: exact synchronization of the two video circuits and the
' fact that all COGs' driven output states get OR'd together, allowing one COG
' to output lows during its preparatory state while the other COG effectively
' drives the pins to create the visible and sync portions of its scan lines.
' During non-visible scan lines, both COGs output together in unison.
'
' COG RAM usage:        $000    = d0 - used to inc destination fields for indirection
'                $001-$050 = scanbuff - longs which hold 4 scan lines
'                $051-$0a0 = colorbuff - longs which hold colors for 80 characters
'                $0a1-$142 = scancode - stacked WAITVID/SHR for fast display
'                $143-$1EF = maincode - main program loop which drives display

                        org     0                               ' set origin to $000 for start of program

d0                      long    1 << 9                          ' d0 always resides here at $000, executes as NOP


' Initialization code and data - after execution, space gets reused as scanbuff

                        ' Move main program into maincode area

:move                   mov     $1EF, main_begin + main_size - 1
                        sub     :move,d0s0                      ' (do reverse move to avoid overwrite)
                        djnz    main_ctr,#:move

                        ' Build scanbuff display routine into scancode

:waitvid                mov     scancode+0, i0                  ' org   scancode
:shr                    mov     scancode+1, i1                  ' waitvid colorbuff+0, scanbuff+0
                        add     :waitvid, d1                    ' shr   scanbuff+0,#8
                        add     :shr, d1                        ' waitvid colorbuff+1, scanbuff+1
                        add     i0, d0s0                        ' shr   scanbuff+1,#8
                        add     i1, d0                          ' ...
                        djnz    scan_ctr, #:waitvid             ' waitvid colorbuff+cols-1, scanbuff+cols-1

                        mov     scancode+cols*2-1, i2           ' mov   vscl,#hf
                        mov     scancode+cols*2+0, i3           ' waitvid hvsync,#0
                        mov     scancode+cols*2+1, i4           ' jmp   #scanret

                        ' Init I/O registers and sync COGs' video circuits

                        mov     dira, reg_dira                  ' set pin directions
                        mov     dirb, reg_dirb
                        movi    frqa, #(pr / 5) << 2            ' set pixel rate
                        mov     vcfg, reg_vcfg                  ' set video configuration
                        mov     vscl, #1                        ' set video to reload on every pixel
                        waitcnt sync_cnt, colormask             ' wait for start value in cnt, add ~1ms
                        movi    ctra, #%00001_110               ' COGs in sync! enable PLLs now - NCOs locked!
                        waitcnt sync_cnt, #0                    ' wait ~1ms for PLLs to stabilize - PLLs locked!
                        mov     vscl, #100                      ' insure initial WAITVIDs lock cleanly

                        ' Jump to main loop

                        jmp     #vsync                          ' jump to vsync - WAITVIDs will now be locked!

                        ' Data

d0s0                    long    1 << 9 + 1
d1                      long    1 << 10
main_ctr                long    main_size
scan_ctr                long    cols

i0                      waitvid colorbuff+0, scanbuff+0
i1                      shr     scanbuff+0, #8
i2                      mov     vscl, #hf
i3                      waitvid hvsync, #0
i4                      jmp     #scanret

reg_dira                long    0                               ' set at runtime
reg_dirb                long    0                               ' set at runtime
reg_vcfg                long    0                               ' set at runtime
sync_cnt                long    0                               ' set at runtime

                        ' Directives

                        fit     scancode                        ' make sure initialization code and data fit
main_begin              org     maincode                        ' main code follows (gets moved into maincode)


' Main loop, display field - each COG alternately builds and displays four scan lines

vsync                   rdlong  col, par                        ' read offset Palette

                        mov     x, #vs                          ' do vertical sync lines
                        call    #blank_vsync

vb_lines                mov     x, #vb                          ' do vertical back porch lines (# set at runtime)
                        call    #blank_vsync

                        mov     screen_ptr, screen_base         ' reset screen pointer to upper-left character
                        mov     row, #0                         ' reset row counter for cursor insertion
                        mov     fours, #rows * 4 / 2            ' set number of 4-line builds for whole screen

                        ' Build four scan lines into scanbuff

fourline                mov     font_ptr, font_part             ' get address of appropriate font section
                        shl     font_ptr, #8+2
                        add     font_ptr, font_base

                        movd    :pixa, #scanbuff-1              ' reset scanbuff address (pre-decremented)
                        movd    :cola, #colorbuff-1             ' reset colorbuff address (pre-decremented)
                        movd    :colb, #colorbuff-1

                        mov     y, #2                           ' must build scanbuff in two sections because
                        mov     vscl, vscl_line2x               ' ..pixel counter is limited to twelve bits

:halfrow                waitvid underscore, #0                  ' output lows to let other COG drive VGA pins
                        mov     x, #cols/2                      ' ..for 2 scan lines, ready for half a row

:column                 rdword  z, screen_ptr                   ' get character and colors from screen memory
                        mov     bg, z
                        and     z, #$ff                         ' mask character code
                        shl     z, #2                           ' * 4
                        add     z, font_ptr                     ' add font section address to point to 8*4 pixels

                        add     :pixa, d0                       ' increment scanbuff destination addresses
                        add     screen_ptr, #2                  ' increment screen memory address

:pixa                   rdlong  scanbuff, z                     ' read pixel long (8*4) into scanbuff
                        ror     bg, #12                         ' background color in bits 3..0
                        mov     fg, bg                          ' foreground color in bits 31..28
                        shr     fg, #28                         ' bits 3..0
                        add     fg, #fg_clut                    ' + offset to foreground CLUT
                        add     fg, col                         ' + offset second Palette

                        movs    :cola, fg
                        add     :cola, d0
                        add     bg, #bg_clut                    ' + offset to background CLUT
                        add     bg, col                         ' + offset second Palette

                        movs    :colb, bg
                        add     :colb, d0

:cola                   mov     colorbuff, 0-0
:colb                   or      colorbuff, 0-0

                        djnz    x, #:column                     ' another character in this half-row?

                        djnz    y, #:halfrow                    ' loop to do 2nd half-row, time for 2nd WAITVID

                        sub     screen_ptr, #2*cols             ' back up to start of same row in screen memory

                        ' Insert cursors into scanbuff

                        mov     z, #2                           ' ready for two cursors

:cursor                 rdbyte  x, cursor_base                  ' x in range?
                        add     cursor_base, #1
                        cmp     x, #cols        wc

                        rdbyte  y, cursor_base                  ' y match?
                        add     cursor_base, #1
                        cmp     y, row          wz

                        rdbyte  y, cursor_base                  ' get cursor mode
                        add     cursor_base, #1

        if_nc_or_nz     jmp     #:nocursor                      ' if cursor not in scanbuff, no cursor

                        add     x, #scanbuff                    ' cursor in scanbuff, set scanbuff address
                        movd    :xor, x

                        test    y, #%010        wc              ' get mode bits into flags
                        test    y, #%001        wz
        if_nc_and_z     jmp     #:nocursor                      ' if cursor disabled, no cursor

        if_c_and_z      test    slowbit, cnt    wc              ' if blink mode, get blink state
        if_c_and_nz     test    fastbit, cnt    wc

                        test    y, #%100        wz              ' get box or underscore cursor piece
        if_z            mov     x, longmask
        if_nz           mov     x, underscore
        if_nz           cmp     font_part, #3   wz              ' if underscore, must be last font section

:xor    if_nc_and_z     xor     scanbuff, x                     ' conditionally xor cursor into scanbuff

:nocursor               djnz    z, #:cursor                     ' second cursor?

                        sub     cursor_base, #3*2               ' restore cursor base

                        ' Display four scan lines from scanbuff

                        mov     y, #4                           ' ready for four scan lines

scanline                mov     vscl, vscl_chr                  ' set pixel rate for characters
                        jmp     #scancode                       ' jump to scanbuff display routine in scancode
scanret                 mov    vscl, #hs                        ' do horizontal sync pixels
                        waitvid hvsync, #1                      ' #1 makes hsync active
                        mov     vscl, #hb                       ' do horizontal back porch pixels
                        waitvid hvsync, #0                      ' #0 makes hsync inactive
                        shr     scanbuff+cols-1, #8             ' shift last column's pixels right by 8
                        djnz    y, #scanline                    ' another scan line?

                        ' Next group of four scan lines

                        add     font_part, #2                   ' if font_part + 2 => 4, subtract 4 (new row)
                        cmpsub  font_part, #4           wc      ' c=0 for same row, c=1 for new row
        if_c            add     screen_ptr, #2*cols             ' if new row, advance screen pointer
        if_c            add     row, #1                         ' if new row, increment row counter
                        djnz    fours, #fourline                ' another 4-line build/display?

                        ' Visible section done, do vertical sync front porch lines

vf_lines                mov     x,#vf                           ' do vertical front porch lines (# set at runtime)
                        call    #blank

                        jmp     #vsync                          ' new field, loop to vsync

                        ' Subroutine - do blank lines

blank_vsync             xor     hvsync,#$101                    ' flip vertical sync bits

blank                   mov     vscl, hx                        ' do blank pixels
                        waitvid hvsync, #0
                        mov     vscl, #hf                       ' do horizontal front porch pixels
                        waitvid hvsync, #0
                        mov     vscl, #hs                       ' do horizontal sync pixels
                        waitvid hvsync, #1
                        mov     vscl, #hb                       ' do horizontal back porch pixels
                        waitvid hvsync, #0
                        djnz    x,#blank                        ' another line?
blank_ret
blank_vsync_ret
                        ret

                        ' Data

screen_base             long    0                               ' set at runtime (3 contiguous longs)
cursor_base             long    0                               ' set at runtime
font_base               long    0                               ' set at runtime

font_part               long    0                               ' set at runtime

hx                      long    hp                              ' visible pixels per scan line
vscl_line               long    hp + hf + hs + hb               ' total number of pixels per scan line
vscl_line2x             long    (hp + hf + hs + hb) * 2         ' total number of pixels per 2 scan lines
vscl_chr                long    1 << 12 + 8                     ' 1 clock per pixel and 8 pixels per set
colormask               long    $FCFC                           ' mask to isolate R,G,B bits from H,V
longmask                long    $FFFFFFFF                       ' all bits set
slowbit                 long    1 << 24                         ' cnt mask for slow cursor blink
fastbit                 long    1 << 23                         ' cnt mask for fast cursor blink

underscore              long    $FFFF0000                       ' underscore cursor pattern

hv                      long    hv_inactive                     ' -H,-V states
hvsync                  long    hv_inactive ^ $200              ' +/-H,-V states


bg_clut                 long    %00000011_00000011              ' black           0
                        long    %00000011_00000111              ' dark blue       1
                        long    %00000011_00010011              ' dark green      2
                        long    %00000011_00010111              ' dark cyan       3
                        long    %00000011_01000011              ' dark red        4
                        long    %00000011_01000111              ' dark magenta    5
                        long    %00000011_10010011              ' light orange    6
                        long    %00000011_11111111              ' white           7
                        long    %00000011_11110011              ' light yellow    8
                        long    %00000011_01010111              ' gray            9

bg_second               long    %00000011_00000011              ' black           0
                        long    %00000011_11001111              ' light magenta   1
                        long    %00000011_11000011              ' light red       2
                        long    %00000011_10010011              ' light orange    3
                        long    %00000011_00110011              ' light green     4
                        long    %00000011_00111111              ' light cyan      5
                        long    %00000011_00001111              ' light blue      6
                        long    %00000011_11111111              ' white           7
                        long    %00000011_00000111              ' dark blue       8
                        long    %00000011_00000111              ' dark blue       9
                        long    %00000011_00000111              ' dark blue      10
                        long    %00000011_00010111              ' dark cyan      11
                        long    %00000011_01000111              ' dark magenta   12
                        long    %00000011_01001011              ' light purple   13
                        long    %00000011_11110011              ' light yellow   14
                        long    %00000011_01010111              ' gray           15


fg_clut                 long    %00000011_00000011              ' black           0
                        long    %00000111_00000011              ' dark blue       1
                        long    %00010011_00000011              ' dark green      2
                        long    %00010111_00000011              ' dark cyan       3
                        long    %01000011_00000011              ' dark red        4
                        long    %01000111_00000011              ' dark magenta    5
                        long    %10010011_00000011              ' light orange    6
                        long    %11111111_00000011              ' white           7
                        long    %11110011_00000011              ' light yellow    8
                        long    %01010111_00000011              ' gray            9

fg_second               long    %00000011_00000011              ' black           0
                        long    %11001111_00000011              ' light magenta   1
                        long    %11000011_00000011              ' light red       2
                        long    %10010011_00000011              ' light orange    3
                        long    %00110011_00000011              ' light green     4
                        long    %00111111_00000011              ' light cyan      5
                        long    %00001111_00000011              ' light blue      6
                        long    %11111111_00000011              ' white           7
                        long    %00000111_00000011              ' dark blue       8
                        long    %00000111_00000011              ' dark blue       9
                        long    %00000111_00000011              ' dark blue      10
                        long    %00010111_00000011              ' dark cyan      11
                        long    %01000111_00000011              ' dark magenta   12
                        long    %01001011_00000011              ' light purple   13
                        long    %11110011_00000011              ' light yellow   14
                        long    %01010111_00000011              ' gray           15



                        ' Uninitialized data

screen_ptr              res     1
font_ptr                res     1

x                       res     1
y                       res     1
z                       res     1
fg                      res     1
bg                      res     1

row                     res     1
fours                   res     1

col                     res     1

                        fit     $1f0





{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                   TERMS OF USE: Parallax Object Exchange License                                             |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}